.. Document meta section

:orphan:

.. Document body

.. Anchors

.. _ansible_collections.@{plugin_name}@_@{plugin_type}@:

.. Title

{% set title = plugin_name -%}

@{ title }@
@{ '+' * title|length }@


The documentation for the @{ plugin_type }@ plugin, @{ plugin_name }@,  was malformed.

The errors were:

{% for error in nonfatal_errors %}
* ::

@{ error | indent(width=8, first=True) }@

{% endfor %}

File a bug with the `@{ collection }@ collection <https://galaxy.ansible.com/@{collection | replace(".", "/", 1)}@>`_ in order to have it corrected.
