:orphan:

.. _list_of_collections:

Collection Index
================

These are the collections with docs hosted on `docs.ansible.com <https://docs.ansible.com/>`_.

{% for name in collections | sort %}
* :ref:`@{ name }@ <plugins_in_@{ name }@>`
{% endfor %}

{% if breadcrumbs %}
.. toctree::
    :maxdepth: 1
    :hidden:

{%   for namespace in namespaces | sort %}
    @{ namespace }@/index
{%   endfor %}
{% endif %}
