This is the {{ collection_name }} Ansible collection.

You can install this using::

    pip install --user ansible-collections-{{ collection_name }}

Or from https://galaxy.ansible.com/ using::

    ansible-galaxy collection install {{ collection_name }}

Note that versions installed via ansible-galaxy will always take precedence
over versions installed via pip even if the pip version is newer than the
ansible-galaxy version.
